//#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "sega.h"
#include "font.h"
#include "tiles.h"
#include "libio.h"

#include "math.h"

typedef unsigned char  bool;    // i guess bool is a c++ thing
#define FALSE 0
#define TRUE  1

//#define DEBUGTRACE      // debug path tracer
//#define DEBUGCHAIN      // debug chain detect
#define SHOWLIMIT       // show time limit

#define HERTZ       60
#define NROWS       10
#define NCOLS       7
#define CENTER      3
#define DSPBOMS     4       // max number of bombs to display by score
#define LAYWID      (NCOLS + 2)
#define LAYHGT      (NROWS + 2)
#define LAYSIZ      (LAYWID * LAYHGT)
#define TOPLINE     (3*8 + 128)
#define BOTLINE     (TOPLINE+(NROWS*2*8))
#define HSPEED      4
#define BOOMSPEED   5
#define FALLSPEED   1
#define NBOOMS      9       // number of boom flashes
#define SCORWID     5       // width of scores
#define BOMBSC      100     // number of points to get a free bomb
#define GAMEOVERTIME (HERTZ * 4)
#define MAXSTACK    LAYSIZ  // 106 stack may be max that will be used, and LAYSIZ=108
#define TIMELIMIT   120     // level time limit

enum
{
    B_Up    = 1,
    B_Down  = 2,
    B_Left  = 4,
    B_Right = 8,
    B_DIR   = 15,
    B_B     = 16,
    B_C     = 32,
    B_A     = 64,
    B_Start = 128
};


enum
{ // tile types
    _BLANK      = 0,    // first four tiles do not rotate
    _BLANK_SQ   = 1,
    _BOMB       = 2,
    _CROSS      = 3,
    _CURVE      = 4,
    _TJOINT     = 8,
    _CAP        = 12,
    _2CURVE     = 16,
    _STRAIGHT   = 20,
    _CROSSOVER  = 24,
    _BOOM       = 28,   // these don't rotate either
    _BOOM1      = 28,
    _BOOM2      = 29,
    _BUBBLE1    = 30,   // these two tile numbers are reserved for later
    _BUBBLE2    = 31,
};


enum
{ // special font characters
    BLANK   = 0x00,
    CARET   = 0x1B,
    SMBOMB  = 0x1C,
};


enum
{ // bits in layout array
    L_PATH2 = 0x80, // path 2 flag in layout array
    L_PATH1 = 0x40, // path 1 flag in layout array
    L_BOOM  = 0x20, // boom-me flag in layout array
    L_TILE  = 0x1F, // tile number bits in layout array
};


enum
{ // bits in paths1[] and paths2[]
    P_UP    = 0x08, // up    direction in PATHS1/PATHS2
    P_DOWN  = 0x04, // down  direction in PATHS1/PATHS2
    P_LEFT  = 0x02, // left  direction in PATHS1/PATHS2
    P_RIGHT = 0x01, // right direction in PATHS1/PATHS2
};


// table for path 1
u_char paths1[] =
{
    0x00, 0x00, 0x00, 0x0F, // blank, blank, bomb, cross
    0x0A, 0x09, 0x05, 0x06, // curve: UL, UR, LR, LL
    0x07, 0x0E, 0x0B, 0x0D, // T-down, T-left, T-up, T-right
    0x04, 0x02, 0x08, 0x01, // cap: down, left, up, right
    0x0A, 0x09, 0x0A, 0x09, // curve: UL+LR, UR+LL, UL+LR, UR+LL
    0x0C, 0x03, 0x0C, 0x03, // straight: V, H, V, H
    0x0C, 0x0C, 0x0C, 0x0C, // crossovers
    0x00, 0x00, 0x00, 0x00, // booms and bubbles
};


// table for path 2
u_char paths2[] =
{
    0x00, 0x00, 0x00, 0x00, // blank, blank, bomb, cross
    0x00, 0x00, 0x00, 0x00, // single curves
    0x00, 0x00, 0x00, 0x00, // T intersections
    0x00, 0x00, 0x00, 0x00, // caps
    0x05, 0x06, 0x05, 0x06, // curve: UL+LR, UR+LL, UL+LR, UR+LL
    0x00, 0x00, 0x00, 0x00, // straights
    0x03, 0x03, 0x03, 0x03, // crossovers
    0x00, 0x00, 0x00, 0x00, // booms and bubbles
};


#if 0
// table for both paths
u_char paths[] =
{
    0x00, 0x00, 0x00, 0x0F, // blank, blank, bomb, cross
    0x0A, 0x09, 0x05, 0x06, // curve: UL, UR, LR, LL
    0x07, 0x0F, 0x0B, 0x0D, // T-down, T-left, T-up, T-right
    0x04, 0x02, 0x08, 0x01, // cap: down, left, up, right
    0x0F, 0x0F, 0x0F, 0x0F, // curve: UL+LR, UR+LL, UL+LR, UR+LL
    0x0C, 0x03, 0x0C, 0x03, // straight: V, H, V, H
    0x0F, 0x0F, 0x0F, 0x0F, // crossovers
    0x00, 0x00, 0x00, 0x00, // booms and bubbles
};
#endif


u_char tiles[][4] =
{
    {  BLANK  , BLANK  , BLANK  , BLANK   }, // BLANK
    { PLAIN_UL,PLAIN_UR,PLAIN_LL,PLAIN_LR }, // BLANK_SQ
    {  BOMB_UL, BOMB_UR, BOMB_LL, BOMB_LR }, // BOMB
    { CROSS_UL,CROSS_UR,CROSS_LL,CROSS_LR }, // CROSS
    { TRNUL_UL,TRNUL_UR,TRNUL_LL,TRNUL_LR }, // CURVE_UL
    { TRNUR_UL,TRNUR_UR,TRNUR_LL,TRNUR_LR }, // CURVE_UR
    { TRNDR_UL,TRNDR_UR,TRNDR_LL,TRNDR_LR }, // CURVE_DR
    { TRNDL_UL,TRNDL_UR,TRNDL_LL,TRNDL_LR }, // CURVE_DL
    {  T_DN_UL, T_DN_UR, T_DN_LL, T_DN_LR }, // TJOINT_D
    { T_LFT_UL,T_LFT_UR,T_LFT_LL,T_LFT_LR }, // TJOINT_L
    {  T_UP_UL, T_UP_UR, T_UP_LL, T_UP_LR }, // TJOINT_U
    { T_RGT_UL,T_RGT_UR,T_RGT_LL,T_RGT_LR }, // TJOINT_R
    { CAP_D_UL,CAP_D_UR,CAP_D_LL,CAP_D_LR }, // CAP_D
    { CAP_L_UL,CAP_L_UR,CAP_L_LL,CAP_L_LR }, // CAP_L
    { CAP_U_UL,CAP_U_UR,CAP_U_LL,CAP_U_LR }, // CAP_U
    { CAP_R_UL,CAP_R_UR,CAP_R_LL,CAP_R_LR }, // CAP_R
    { UL_DR_UL,UL_DR_UR,UL_DR_LL,UL_DR_LR }, // 2CURVE UL+LR
    { UR_DL_UL,UR_DL_UR,UR_DL_LL,UR_DL_LR }, // 2CURVE UR+LL
    { UL_DR_UL,UL_DR_UR,UL_DR_LL,UL_DR_LR }, // 2CURVE UL+LR
    { UR_DL_UL,UR_DL_UR,UR_DL_LL,UR_DL_LR }, // 2CURVE UR+LL
    {  VERT_UL, VERT_UR, VERT_LL, VERT_LR }, // STRAIGHT V
    { HORIZ_UL,HORIZ_UR,HORIZ_LL,HORIZ_LR }, // STRAIGHT H
    {  VERT_UL, VERT_UR, VERT_LL, VERT_LR }, // STRAIGHT V
    { HORIZ_UL,HORIZ_UR,HORIZ_LL,HORIZ_LR }, // STRAIGHT H
    { CROSH_UL,CROSH_UR,CROSH_LL,CROSH_LR }, // CROSSOVER V bottom
    { CROSV_UL,CROSV_UR,CROSV_LL,CROSV_LR }, // CROSSOVER H bottom
    { CROSH_UL,CROSH_UR,CROSH_LL,CROSH_LR }, // CROSSOVER V bottom
    { CROSV_UL,CROSV_UR,CROSV_LL,CROSV_LR }, // CROSSOVER H bottom
    { STARL_UL,STARL_UR,STARL_LL,STARL_LR }, // BOOM1 big star
    { STARS_UL,STARS_UR,STARS_LL,STARS_LR }, // BOOM2 small star
    { PLAIN_UL,PLAIN_UR,PLAIN_LL,PLAIN_LR }, // (reserved)
    { PLAIN_UL,PLAIN_UR,PLAIN_LL,PLAIN_LR }, // (reserved)
};
// #define NTILES (sizeof tiles/sizeof tiles[0])


enum
{ // game states
    ST_InitLevel,   // initializing level
    ST_Player,      // normal gameplay
    ST_NoPlayer,    // player not playing (never started or game over)
    ST_Over,        // display a GAME OVER message and go to NoPlayer
    ST_AllOver,     // GAME OVER has been displayed
    ST_Boom,        // tiles exploding
    ST_Fall,        // tiles falling
};


typedef struct
{
//  u_short st_mode;        // current mode for each player
    short   st_timr;        // timer for the current state
    u_short st_count;       // counter for the current state

    u_short mapBase;        // start of VRAM for map
    u_short scrBase;        // start of VRAM for score
    u_short sprBase;        // start of VRAM for sprites
    u_short bomBase;        // start of VRAM for bombs
    u_short limBase;        // start of VRAM for time limit
    u_short hvLink;         // HV size/link word from sprite

    u_char  p2Flag;         // 0=P1, 1=P2
    u_char  state;          // current handler for game state

    u_short lftCap;         // caps for left/right edges (low bit = row 1 has a cap)
//  u_short rgtCap;
    u_short botCap;         // caps for bottom edge (low bit = col 1 has a cap)

    u_short vSpeed;         // current vertical speed

    u_short airNum;         // tile number of current "in-air" piece, per player
    u_short nxtNum;         // tile number of next piece, per player
    u_short nxtNum2;        // real next tile (current might be called bomb)

    u_short airLine;        // scan line of current "in-air" piece, per player
    u_short nxtLine;        // scan line of next piece

    u_short lftHorz;        // horizontal pixel of left edge of map

    u_short airCol;         // column number of current "in-air" piece, per player

//  u_short nBoom;          // number of tiles to go boom

    u_short bombs;          // number of bombs left

    u_short limit;          // remaining time for level
    u_short limitTick;      // ticks counter for limit

//  u_short drpFlg;         // flag if tiles dropped in ST_FALL mode
//  u_short drpLay;         // layout pointer in ST_FALL mode
//  u_short drpMap;         // character map pointer in ST_FALL mode

    u_long  score;          // score (binary)

    volatile u_char *joyBase;   // address of joypad input
    u_short joy;            // saved state of joypad
    u_short joyLast;        // last joypad state (for up direction)
    u_short joyOn;          // changed-to-on joypad buttons
    u_short joyCnt;         // timer for horizontal repeat, per player
//  u_short fire;           // saved state of fire button, per player
//  u_short fireCnt;        // timer for fire button repeat, per player
//  u_short vertCnt;        // timer for vertical motion, per player

    union                   // layout data, as 1D and 2D arrays
    {       // bits 7-6 = trace flags, bit 5 = boom-me, bits 4-0 = tile number
//      u_char  lay0ut[LAYHGT][LAYWID];
        u_char  layout[LAYSIZ];
    };
} PlayerRec;


typedef struct
{
    u_char row;     // row of square
    u_char col;     // column of square
    u_char ofs;     // offset into layout for square
    u_char fromDir; // entry direction, same as P_ bits
} StackRec;


//--------------------------------------------
// global variables
//--------------------------------------------
PlayerRec   P1,P2;              // stuff for each player
bool        pauseFlag;          // TRUE if game paused
bool        gameOver;           // TRUE if game over (time to return to main menu)
bool        twoPlayers;         // 0=1 player, 1=2 players
bool        titleExit;          // TRUE if exiting title screen (time to start game)
bool        goodPath;           // TRUE as long as path being traced isn't blocked
bool        pathList[LAYSIZ];   // ofs of tiles in current path (FIXME: can be made into a bitmap)
StackRec    stack[MAXSTACK];    // tracing stack
short       stackPtr;           // top of tracing stack (next free entry)
u_char      buffer[256];        // generic buffer for string conversions, etc.
u_long      hiScore;            // current high score

#ifdef DEBUGCHAIN
bool chainCleared;
#endif

//--------------------------------------------
// Vertical retrace interrupt handler
//--------------------------------------------
int frameCount = 0; // vsync interrupt frame counter

void __attribute__((interrupt_handler)) VSYNC_INTERRUPT(void)
{
    frameCount++;
}


//--------------------------------------------
// Initialize Color RAM
//--------------------------------------------
static inline void InitCRAM(void)
{
    static const short ctab[] =
    {
        // color table 0 = normal text
        0x222,0x000,0x2D2,0x6F6,0xF22,0xF64,0x12B,0xFD4,0x22F,0x66F,0x2DD,0x9DD,0x292,0xB4D,0xBBB,0xFFF,
        // color table 1 = inverse video
        0x222,0xFFF,0x2D2,0x6F6,0xF22,0xF64,0x12B,0xFD4,0x22F,0x66F,0x2DD,0x9DD,0x292,0xB4D,0xBBB,0x000,
        // color table 2 = black
        0x222,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,0x000,
        // color table 3 = dim graphics
        0x222,0x000,0x161,0x373,0x711,0x732,0x015,0x762,0x117,0x337,0x166,0x466,0x141,0x526,0x555,0x777
    };

    int i;
    const short *p;

    p = ctab;
    *VDP_CTRL_L = CRAM_W;
    for (i = sizeof ctab / sizeof ctab[0]; i; i--)
        *VDP_DATA = *p++;
}


//--------------------------------------------
// Wait for a vertical blank
//--------------------------------------------
void WaitVBlank(void)
{
    while (!(*VDP_STAT & 0x08)) /* wait */;
    while   (*VDP_STAT & 0x08)  /* wait */;
}


//--------------------------------------------
// Initialize joypad ports
//--------------------------------------------
void InitJoy(void)
{
    // 0x40 = all bits input, except TH=output
    *CTRL1_CTRL = 0x40;
    *CTRL2_CTRL = 0x40;
    *CTRL3_CTRL = 0x40;
}


//--------------------------------------------
// Return a random tile number
//--------------------------------------------
int RandTile(void)
{
    int tile;

    do
    {
        tile = rand() & L_TILE;
#if 0
    } while (tile < _CROSS || tile >= _BOOM);
#else // it's a pain to test with all those crossovers around
    } while (tile < _CROSS || tile >= _BOOM || tile==_CROSSOVER || tile==_CROSS
                || tile==_STRAIGHT || tile == _TJOINT);
#endif
    return tile;
}


//--------------------------------------------
// Clear the screen to all BLANK characters
//--------------------------------------------
#define NAMEOFS(row,col) ((row)*0x80 + (col)*2)
#define NAMEPOS(row,col) (NAMEOFS(row,col)<<16)

void ClearScreen(void)
{
    int row,col;

    for (row=0; row<28; row++)
    {
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(row,0);
        for (col=0; col<40; col++)
            *VDP_DATA = BLANK;
    }
}


//--------------------------------------------
// Copy a string to the VDP
//--------------------------------------------
void ScrnMsg(char *s)
{
    while (*s) *VDP_DATA = *s++;
}


void hex1(int i)
{
    i = i & 15;
    i = i + '0';
    if (i>'9') i=i+7;
    *VDP_DATA = i;
}


void hex2(int i)
{
    hex1(i>>4);
    hex1(i&15);
}


void hex4(int i)
{
    hex2(i>>8);
    hex2(i&255);
}


void hex8(long i)
{
    hex4(i>>16);
    hex4(i&65535);
}


#define DEBUGTOP 5
#define DEBUGLEFT 18
#define DEBUGRIGHT 40

short debugRow,debugCol;

void DebugHome(void)
{
    debugRow = DEBUGTOP;
    debugCol = DEBUGLEFT;
}


void DebugClear(void)
{
    int row,col;

    DebugHome();

    for (row=0; row<28; row++)
    {
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(row,DEBUGLEFT);
        for (col=DEBUGLEFT; col<40; col++)
            *VDP_DATA = BLANK;
    }
}


void DebugNL(void)
{
    debugRow++;
    if (debugRow>27) debugRow=DEBUGTOP;
    debugCol = DEBUGLEFT;
}


void DebugChar(u_char c)
{
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(debugRow,debugCol);
    *VDP_DATA = c;

    if (debugCol >= DEBUGRIGHT)
        DebugNL();
    debugCol++;
    if (debugCol >= DEBUGRIGHT)
        DebugNL();
}


void DebugHex1(int i)
{
    i = i & 15;
    i = i + '0';
    if (i>'9') i=i+7;
    DebugChar(i);
}


void DebugHex2(int i)
{
    DebugHex1(i>>4);
    DebugHex1(i&15);
}


void DebugHex4(int i)
{
    DebugHex2(i>>8);
    DebugHex2(i&255);
}


void DebugHex8(long i)
{
    DebugHex4(i>>16);
    DebugHex4(i&65535);
}


void DebugStr(u_char *s)
{
    while(*s) DebugChar(*s++);
}


//--------------------------------------------
// Show joypad input for one player
//--------------------------------------------
void ShowJoy(PlayerRec *pl)
{
    char* name = "SACBRLDU";
    int port;
    int i,n,c;

    for (port = 0; port <= 1; port++)
    {
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(0,pl->p2Flag*20);

        n = pl->joy;
        for (i = 7; i >= 0; i--)
        {
            c = '-';
            if (n & (1 << i)) c = name[7 - i];
            *VDP_DATA = c;
        }
    }
}


//--------------------------------------------
// Read joypad input for one player
//--------------------------------------------
void RdJoy(PlayerRec *pl)
{
    u_char i,j;

    pl->joyLast = pl->joy;      // update last joypad state

    *pl->joyBase = 0x40;        // turn on TH output
    asm volatile("nop;nop");    // delay a moment
    i = *pl->joyBase;           // read controller port (--CBRLDU)
    i = i | ~0x3F;              // mask off unused bits

    *pl->joyBase = 0x00;        // turn off TH output
    asm volatile("nop;nop");    // delay a moment
    j = *pl->joyBase;           // read controller port (--SA----)
    j = j << 2;                 // shift TH=0 bits left (SA------)
    j = j | 0x3F;               // mask off unused bits

    pl->joy = ~ (i & j);

    if (((pl->joyLast & B_DIR) == 0) ||     // always reset count if none pressed
        (pl->joy ^ pl->joyLast) & B_DIR)    // has the direction changed?
    {
        pl->joyCnt = 0;         // reset timer if changed
    }

    pl->joyOn = (pl->joy ^ pl->joyLast) & pl->joy;

#if 0
    ShowJoy(pl);
#endif
}


//--------------------------------------------
//
//--------------------------------------------
void halt(char *s)
{
    int i;

    *VDP_CTRL_L = NAMEA_W + NAMEPOS(27,0);
    ScrnMsg(s);

    do RdJoy(&P1); while((P1.joyOn & B_A) == 0);
    
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(27,0);
    for (i=0; i<40; i++)
        *VDP_DATA = 0;
}


//--------------------------------------------
// Randomize the random number generator
//--------------------------------------------
void Randomize(void)
{
    srand(frameCount);
}


//--------------------------------------------
// Initialize the data for each player
//--------------------------------------------
void InitPlayers(void)
{
    pauseFlag = FALSE;

    PlayerRec *pl;
    int i;

    pl = &P1;
    for (i=0; i<2; i++)
    {
        memset(pl,0,sizeof *pl);
        pl->vSpeed = 15;
#if 0
        pl->lftCap = 0x03FF;
        //pl->rgtCap = 0x03FF;
        pl->botCap = 0x007F;
#else
        pl->botCap = 0;
        pl->lftCap = 0x0155;
        //pl->rgtCap = 0x0155;
        pl->botCap = 0x00AA;
#endif
        pl->airLine = TOPLINE;
        pl->airCol = CENTER;
        pl->score = 0;
//pl->score = 55555;
        pl->limit = TIMELIMIT;
        pl->limitTick = 0;

        pl->state = ST_InitLevel;

        if (i==0 || twoPlayers)
        {
            pl->airNum  = RandTile();
            pl->nxtNum  = RandTile();
            pl->nxtNum2 = pl->nxtNum;
            pl->bombs   = 4;
        }

        pl = &P2;
    }

    P1.joyBase = CTRL1_DATA;
    P2.joyBase = CTRL2_DATA;
    P2.p2Flag = 1;
}


//--------------------------------------------
// Clear the sprite attribute table
//--------------------------------------------
void ClearSprites(void)
{
    *VDP_CTRL_L = SPRTATTR_W;
    *VDP_DATA_L = 0; // clear Y offset, HV size, link
    *VDP_DATA_L = 0; // clear Y offset, HV size, link
}


//--------------------------------------------
// Initialize sprite attribute table
//--------------------------------------------
void InitSprites(void)
{
    *VDP_CTRL_L = SPRTATTR_W;

    *VDP_DATA = 0;          // 0x00 Y offset P1
    *VDP_DATA = 0x0501;     // 0x02 HV size, link
    *VDP_DATA = 0;          // 0x04 cell
    *VDP_DATA = 0;          // 0x06 X offset

    *VDP_DATA = 2*8+128;    // 0x08 Y offset P1 next
    *VDP_DATA = 0x0502;     // 0x0A HV size, link
    *VDP_DATA = 0;          // 0x0C cell
    *VDP_DATA = 17*8+4+128; // 0x0E X offset

    *VDP_DATA = 0;          // 0x10 Y offset P2
    *VDP_DATA = 0x0503;     // 0x12 HV size, link
    *VDP_DATA = 0;          // 0x14 cell
    *VDP_DATA = 0;          // 0x16 X offset

    *VDP_DATA = 2*8+128;    // 0x18 Y offset P2 next
    *VDP_DATA = 0x0500;     // 0x1A HV size, link
    *VDP_DATA = 0;          // 0x1C cell
    *VDP_DATA = 21*8-4+128; // 0x1E X offset

    P1.sprBase = 0;
    P2.sprBase = 16;

    P1.mapBase = NAMEOFS(5,2);
    P2.mapBase = NAMEOFS(5,2+20);

    P1.scrBase = NAMEOFS(2,7);
    P2.scrBase = NAMEOFS(2,7+20);

    P1.limBase = NAMEOFS(2,14);
    P2.limBase = NAMEOFS(2,3+20);

    P1.bomBase = NAMEOFS(2,8-6);
    P2.bomBase = NAMEOFS(2,8+20+6);

    P1.hvLink = 0x0501;
    P2.hvLink = 0x0503;

    P1.lftHorz =  3*8 + 128;
    P2.lftHorz = 23*8 + 128;
}

//  Index + 0  :   ------yy yyyyyyyy Y offset + 128
//  Index + 2  :   ----hhvv          H,V size (0=1, 1=2, 2=3, 3=4)
//  Index + 3  :   -lllllll          link to next sprite
//  Index + 4  :   pccvhnnn nnnnnnnn priority, palette, flip, starting cell
//  Index + 6  :   ------xx xxxxxxxx X offset + 128
// spirte 0 = player 1 current in-air tile
// sprite 1 = player 1 next tile
// sprite 2 = player 2 current in-air tile
// spirte 3 = player 2 next tile

//--------------------------------------------
// Update sprites for falling tiles and next tiles
//--------------------------------------------
void UpdSprite(PlayerRec *pl)
{
    long base;

    // compute sprite base address
    base = pl->sprBase;
    base = SPRTATTR_W + (base<<16);

    // update falling tile
    *VDP_CTRL_L = base;
    *VDP_DATA = pl->airLine;                    // Y offset
    *VDP_DATA = pl->hvLink;                     // HV size, link
    *VDP_DATA = tiles[pl->airNum][0];           // starting cell number
    *VDP_DATA = pl->airCol*16 + pl->lftHorz;    // X offset

    // update next tile
    *VDP_CTRL_L = base + (0x0C<<16);  // point to cell number of second sprite

    // update next tile
    *VDP_DATA = tiles[pl->nxtNum][0];           // starting cell number
}


//--------------------------------------------
// 
//--------------------------------------------
void UpdScore(PlayerRec *pl)
{
    int i,n;

    if (pl->p2Flag && !twoPlayers)
        return;

    // FIXME: make dedicated bin->asc converter
    sprintf(buffer,"%06lu",pl->score);    // convert player's score to ASCII
    *VDP_CTRL_L = NAMEA_W + (pl->scrBase << 16);    // set VDP address to player score
    ScrnMsg(buffer);

    n = pl->bombs;      // get number of bombs
    if (n>4) n=4;       // show 4 max

    *VDP_CTRL_L = NAMEA_W + (pl->bomBase << 16);    // set VDP address to player's bombs
    n = pl->bombs;              // get number of bombs
    if (n>DSPBOMS) n=DSPBOMS;   // limit number of on-screen bombs

    if (pl->p2Flag)             // if player 2, display blanks on left
    {
        for (i=0; i<DSPBOMS-n; i++)
            *VDP_DATA = BLANK;
        for (i=0; i<n; i++)
            *VDP_DATA = SMBOMB;
    }
    else
    {
        for (i=0; i<n; i++)
            *VDP_DATA = SMBOMB;
        for (i=0; i<DSPBOMS-n; i++)
            *VDP_DATA = BLANK;
    }

#ifdef SHOWLIMIT
    *VDP_CTRL_L = NAMEA_W + (pl->limBase << 16);    // set VDP address to player time limit
    sprintf(buffer,"%3u", pl->limit);
    ScrnMsg(buffer);
#endif
}


//--------------------------------------------
// Add n to player's score 
//--------------------------------------------
void AddScore(PlayerRec *pl, int n)
{
    pl -> score += n;
    UpdScore(pl);
}


//--------------------------------------------
// 
//--------------------------------------------
void DrawMap(PlayerRec *pl)
{
    int edge,row,col,tile;
    long va;
    u_char *p;

    edge = pl->lftCap;        // vertical edges
    p = pl->layout + 1 + LAYWID;    // map pointer

    va = NAMEA_W + (pl->mapBase<<16); // VRAM address of start of map

    for (row=0; row<NROWS; row++)   // row = 0 .. NROWS-1
    {
        // display higher half
        *VDP_CTRL_L = va;       // VRAM address = first line

        if (row == 0)
        {
            if (edge & 1)
                *VDP_DATA = TOPLFT_C;
            else
                *VDP_DATA = TOPLFT;
        }
        else
        {
            if (edge & 1)
                *VDP_DATA = LFT_C_U;
            else
                *VDP_DATA = SIDE;
        }

        for (col=0; col<NCOLS; col++)   // col = 0 .. NCOLS-1
        {
            tile = *p++ & L_TILE;
            *VDP_DATA = tiles[tile][0];
            *VDP_DATA = tiles[tile][1];
        }

        if (row == 0)
        {
            if (edge & 1)
                *VDP_DATA = TOPRGT_C;
            else
                *VDP_DATA = TOPRGT;
        }
        else
        {
            if (edge & 1)
                *VDP_DATA = RGT_C_U;
            else
                *VDP_DATA = SIDE;
        }

        // display lower half
        va += 0x80<<16;         // move to next VRAM row
        *VDP_CTRL_L = va;       // VRAM address = second line

        if (edge & 1)           // put left edge in VRAM
            *VDP_DATA = LFT_C_L;
        else
            *VDP_DATA = SIDE;

        p -= NCOLS;             // return to beginning of row
        for (col=0; col<NCOLS; col++)   // col = 0 .. NCOLS-1
        {
            tile = *p++ & L_TILE;
            *VDP_DATA = tiles[tile][2];
            *VDP_DATA = tiles[tile][3];
        }
        p += LAYWID-NCOLS;      // skip the padding in the map

        if (edge & 1)           // put right edge in VRAM
            *VDP_DATA = RGT_C_L;
        else
            *VDP_DATA = SIDE;

        edge >>= 1;             // shift to next edge bit
        va += 0x80<<16;         // move to next VRAM row
    }

    // display bottom edge
    edge = pl->botCap;          // horizontal edges
    *VDP_CTRL_L = va;           // VRAM address = bottom line
    *VDP_DATA = BOTLFT;         // put lower left corner in VRAM

    for (col=0; col<NCOLS; col++)   // col = 0 .. NCOLS-1
    {
        if (edge & 1)           // test next bit of bottom edge
            *VDP_DATA_L = (BOT_C_L<<16) + BOT_C_R;
        else
            *VDP_DATA_L = (BOT<<16) + BOT;
        edge >>= 1;             // shift to next edge bit
    }

    *VDP_DATA = BOTRGT;         // put lower right corner in VRAM
}

//--------------------------------------------
// Check for current player pausing the game
//--------------------------------------------
void DoPause(PlayerRec *pl)
{
//  if (pl->p2Flag && !twoPlayers) return;  // don't check pause if P2 && 1P

    if (pl->joyOn & B_Start)       // test start button
    {
        pauseFlag = TRUE;       // pause the game
        if (P1.state != ST_NoPlayer)
        {
            *VDP_CTRL_L = NAMEA_W + (P1.mapBase<<16) + NAMEPOS(9,4);
            ScrnMsg(" PAUSED ");
        }

        if (P2.state != ST_NoPlayer)
        {
            *VDP_CTRL_L = NAMEA_W + (P2.mapBase<<16) + NAMEPOS(9,4);
            ScrnMsg(" PAUSED ");
        }
    }
}


//--------------------------------------------
// 
//--------------------------------------------
void DoLeftRight(PlayerRec *pl)
{
    int row,col;
    u_char *p;

    if (pl->joyCnt != 0)    // has joypad changed or timed out?
    {
        pl->joyCnt--;       // if not, decrement timer and return
        return;
    }

    if (pl->joy & B_Left)           // left
        col = -1;
    else if (pl->joy & B_Right)     // right
        col =  1;
    else return;

    col += pl->airCol;              // add left/right offset
    if (col < 0 || col >= NCOLS)    // exit if already at edge
        return;

    row = pl->airLine - TOPLINE;     // how far from the top of the map are we?

    p = pl->layout + ((row/16) * LAYWID + col + 1); // get address of new space
    if (*p) return;                 // exit if something in new space

    if ((row & 0x0F) == 0 || *(p+LAYWID) == 0)
    {   // if at even tile boundary or row empty
        pl->joyCnt = HSPEED;        // reset delay timer
        pl->airCol = col;           // update position
        UpdSprite(pl);              // update sprite
    }
}


//--------------------------------------------
// 
//--------------------------------------------
void DoButtons(PlayerRec *pl)
{
    int n;

    // handle down button
    if (pl->joyOn & B_Down)
    {
        pl->st_timr = 0;    // force it to start moving down
    }

    // handle up button
    if ((pl->joyOn & B_Up) && pl->bombs)
    {
        // if next tile currently a bomb, get "real" next tile
        if (pl->nxtNum == _BOMB)
            n = pl->nxtNum2;
        else
            n = _BOMB;
//FIXME: aren't these two backwards?
        UpdSprite(pl);      // update next sprite display
        pl->nxtNum = n;     // update next tile
    }

    // handle C button
    if (!(pl->joyOn & B_C)) return; // exit if C button not pressed

    n = pl->airNum & 0x1C;      // get current tile value, mask off direction
    if (n == 0 || n == 0x1C)
        return;                 // can't rotate 0x00-0x03 or 0x1C-0x1F

    pl->airNum = n + ((pl->airNum+1) & 0x03);   // increment tile type

    UpdSprite(pl);              // update sprite to show new tile type
}


//--------------------------------------------
// Clear path tracing flags from all tiles
//--------------------------------------------
void ClrPaths(PlayerRec *pl)
{
    int i;

    for (i=0; i<LAYSIZ; i++)
    {
        pl->layout[i] &= L_TILE;
        pathList[i] = FALSE;
    }
}


//--------------------------------------------
// Returns triangle number of n
//--------------------------------------------
int Triangle(int n)
{
    return (n*n+n)/2;
}


//--------------------------------------------
// Clear path tracing flags from all tiles
//--------------------------------------------
void ClearPaths(PlayerRec *pl)
{
    int i;

    for (i=0; i<sizeof pl->layout; i++)
        pl->layout[i] &= L_TILE;
}


//--------------------------------------------
// Set boom flags for current path
//--------------------------------------------
void BoomPath(PlayerRec *pl)
{
    int i;

#ifdef DEBUGTRACE
DebugClear();
#endif
    for (i=0; i<LAYSIZ; i++)
    {
        if (pathList[i])
        {
#ifdef DEBUGTRACE
DebugChar(' ');
DebugHex2(i);
#endif
            pl->layout[i] |= L_BOOM;
        }
    }
#ifdef DEBUGTRACE
halt("-BoomPath-");
#endif
}


//--------------------------------------------
// Update one tile on the map
// pos
//  0 = normal
//  1 = down one half row
//  2 = only put blanks in bottom half
// FIXME: need enums for he above
//--------------------------------------------
void UpdTile(PlayerRec *pl, int tile, int row, int col, int pos)
{
    long addr;
    u_char *p;

    addr = NAMEA_W + (pl->mapBase<<16) + NAMEPOS(row*2, col*2+1);
    if (pos)
    {
        addr += (128<<16);

        if (pos != 1)
        {
            *VDP_CTRL_L = addr;
            *VDP_DATA = BLANK;  // store two blanks
            *VDP_DATA = BLANK;
            return;
        }
    }
    *VDP_CTRL_L = addr;     // screen address of first line

    tile &= L_TILE;         // mask off flag bits
    p = tiles[tile];        // get cells for this tile

    *VDP_DATA = *p++;       // copy first cell
    *VDP_DATA = *p++;       // copy second cell

    *VDP_CTRL_L = addr + (128<<16);   // screen address of second line

    *VDP_DATA = *p++;       // copy third cell
    *VDP_DATA = *p++;       // copy fourth cell
}


//--------------------------------------------
// Change current tile to _BOOM1 if X,Y are in range and tile not blank
//--------------------------------------------
int PutBoom(PlayerRec *pl, char *p, int row, int col)
{
    if (row<0 || row>=NROWS) return 0;  // return if row out of range
    if (col<0 || col>=NCOLS) return 0;  // return if col out of range
    if (*p == _BLANK) return 0;         // return if square is blank

    *p = _BOOM1;                        // put _BOOM1 in layout
    UpdTile(pl,_BOOM1,row,col,0);
    return 1;
}


//--------------------------------------------
// Convert marked tiles to booms
//--------------------------------------------
void BoomPaths(PlayerRec *pl)
{
    u_char *p;
    int row,col;
#ifdef DEBUGTRACE
DebugClear();
#endif
    p = pl->layout + LAYWID + 1;    // start of map data

    for (row=0; row<NROWS; row++)   // row = NROWS-1..0
    {
        for (col=0; col<NCOLS; col++)   // col = NCOLS-1..0
        {
            if (*p & L_BOOM)        // if tile needs to be boomed
            {
#ifdef DEBUGTRACE
DebugHex1(row);
DebugChar(':');
DebugHex1(col);
DebugChar(' ');
#endif
                PutBoom(pl,p,row,col);  // change tile to a boom
            }
            p++;                    // point to next square
        }
        p += LAYWID - NCOLS;        // go to next map line
    }

    pl->state = ST_Boom;            // change next state to flash the booms
    pl->st_timr = BOOMSPEED;
    pl->st_count = NBOOMS;
#ifdef DEBUGTRACE
halt("-BOOM-");
#endif
}


int maxStack = 0;
//--------------------------------------------
// Trace one path from a tile
//--------------------------------------------
int TracePath(PlayerRec *pl, u_char *sq, int row, int col, int traceFirst)
{
    int     ofs;        // offset of current square relative to pl->layout
    int     exits;      // exits from current square
    bool    firstTime;  // first time through (don't look at stack)
    int     pathLen;

    goodPath  = TRUE;   // path is good until we find out otherwise
    stackPtr  = 0;      // clear trace stack
    pathLen   = 0;
    firstTime = TRUE;   // first time through (don't look at stack)
    ofs = sq - pl->layout;  // get offset of starting square
#ifdef DEBUGTRACE
DebugClear();
#endif
    // while stack not empty
    while (stackPtr != 0 || firstTime)
    {
        if (firstTime)
        {
#ifdef DEBUGTRACE
DebugHex1(row);
DebugChar(':');
DebugHex1(col);
DebugChar(':');
DebugHex2(ofs);
DebugChar('B'-traceFirst);
#endif
            firstTime = FALSE;

            // get exits from current path
            if (traceFirst) exits = paths1[*sq & L_TILE];
                       else exits = paths2[*sq & L_TILE];
//DebugHex1(exits);
        }
        else
        {
            // get last stack entry
            stackPtr--;
            ofs = stack[stackPtr].ofs;
            sq  = pl->layout + ofs;
            row = stack[stackPtr].row;
            col = stack[stackPtr].col;
#ifdef DEBUGTRACE
DebugChar(' ');
DebugHex1(row);
DebugChar(':');
DebugHex1(col);
DebugChar(':');
DebugHex2(ofs);
#endif
            // determine which path we are on
            traceFirst = TRUE;
            if (paths2[*sq & L_TILE] & stack[stackPtr].fromDir)
                traceFirst = FALSE;
            else if (!(paths1[*sq & L_TILE] & stack[stackPtr].fromDir))
            {
                // path does not trace back
#ifdef DEBUGTRACE
                DebugChar('x');
#endif
                goodPath = FALSE;
                continue;
            }
#ifdef DEBUGTRACE
DebugChar('B'-traceFirst);
#endif
            // ignore this square if path already traced
            if (traceFirst)
            {
                if (*sq & L_PATH1)
                {
#ifdef DEBUGTRACE
                    DebugChar('!');
#endif
                    continue;
                }
            }
            else
            {
                if (*sq & L_PATH2)
                {
#ifdef DEBUGTRACE
                    DebugChar('!');
#endif
                    continue;
                }
            }

            // get exits from current path
            if (traceFirst) exits = paths1[*sq & L_TILE];
                       else exits = paths2[*sq & L_TILE];

#ifdef DEBUGTRACE
if (ofs == 0x53)
{
 DebugStr(" X=");
 DebugHex2(exits);
 DebugChar(':');
}
#endif
            // remove incoming direction from exits
            exits &= ~stack[stackPtr].fromDir;
#ifdef DEBUGTRACE
if (ofs == 0x53)
{
 DebugHex2(exits);
}
#endif

//DebugHex1(exits);
        }

        // add current ofs to pathList
        pathList[ofs] = TRUE;
        pathLen++;

        // set appropriate path traced flag
        if (traceFirst)
        {
#ifdef DEBUGTRACE
            DebugChar('a');
#endif
            *sq |= L_PATH1;
        }
        else
        {
#ifdef DEBUGTRACE
            DebugChar('b');
#endif
            *sq |= L_PATH2;
        }

        // for each exit in current path,
        // add fromdir and row/col/ofs of next square to each stack entry

        if (exits & P_UP)
        {
#ifdef DEBUGTRACE
DebugChar('u');
#endif
            // check for top edge of map
            if (row <= 0)
            {
                goodPath = FALSE;
#ifdef DEBUGTRACE
                DebugChar('-');
#endif
            }
            else
            {
#ifdef DEBUGTRACE
DebugChar('[');
DebugHex1(row-1);
DebugChar(':');
DebugHex1(col);
DebugChar(':');
DebugHex2(ofs-LAYWID);
#endif
                stack[stackPtr].ofs = ofs - LAYWID;
                stack[stackPtr].row = row - 1;
                stack[stackPtr].col = col;
                stack[stackPtr].fromDir = P_DOWN;
                stackPtr++;
if (stackPtr>maxStack) maxStack = stackPtr;
#ifdef DEBUGTRACE
DebugChar(']');
#endif
            }
        }

        if (exits & P_DOWN)
        {
#ifdef DEBUGTRACE
DebugChar('d');
#endif
            // check for bottom edge of map
            if (row+1 >= NROWS) 
            {
                if (!(pl->botCap & (1<<col)))
                {
#ifdef DEBUGTRACE
                    DebugChar('-');
#endif
                    goodPath = FALSE;
                }
#ifdef DEBUGTRACE
                else DebugChar('+');
#endif
            }
            else
            {
#ifdef DEBUGTRACE
DebugChar('[');
DebugHex1(row+1);
DebugChar(':');
DebugHex1(col);
DebugChar(':');
DebugHex2(ofs+LAYWID);
#endif
                stack[stackPtr].ofs = ofs + LAYWID;
                stack[stackPtr].row = row + 1;
                stack[stackPtr].col = col;
                stack[stackPtr].fromDir = P_UP;
                stackPtr++;
#ifdef DEBUGTRACE
DebugChar(']');
#endif
            }
        }

        if (exits & P_LEFT)
        {
#ifdef DEBUGTRACE
DebugChar('l');
#endif
            // check for left edge of map
            if (col <= 0)
            {
                if (!(pl->lftCap & (1<<row)))
                {
#ifdef DEBUGTRACE
                    DebugChar('-');
#endif
                    goodPath = FALSE;
                }
#ifdef DEBUGTRACE
                else DebugChar('+');
#endif
            }
            else
            {
#ifdef DEBUGTRACE
DebugChar('[');
DebugHex1(row);
DebugChar(':');
DebugHex1(col-1);
DebugChar(':');
DebugHex2(ofs-1);
#endif
                stack[stackPtr].ofs = ofs - 1;
                stack[stackPtr].row = row;
                stack[stackPtr].col = col - 1;
                stack[stackPtr].fromDir = P_RIGHT;
                stackPtr++;
#ifdef DEBUGTRACE
DebugChar(']');
#endif
            }
        }

        if (exits & P_RIGHT)
        {
#ifdef DEBUGTRACE
DebugChar('r');
#endif
            // check for right edge of map
            if (col+1 >= NCOLS)
            {
                if (!(pl->lftCap & (1<<row))) // rgtCap
                {
#ifdef DEBUGTRACE
                    DebugChar('-');
#endif
                    goodPath = FALSE;
                }
#ifdef DEBUGTRACE
                else DebugChar('+');
#endif
            }
            else
            {
#ifdef DEBUGTRACE
DebugChar('[');
DebugHex1(row);
DebugChar(':');
DebugHex1(col+1);
DebugChar(':');
DebugHex2(ofs+1);
#endif
                stack[stackPtr].ofs = ofs + 1;
                stack[stackPtr].row = row;
                stack[stackPtr].col = col + 1;
                stack[stackPtr].fromDir = P_LEFT;
                stackPtr++;
#ifdef DEBUGTRACE
DebugChar(']');
#endif
            }
        }
    }

    return pathLen;
}


//--------------------------------------------
// Record chain in a bucket
//--------------------------------------------
void AddBucket(PlayerRec *pl, int pathLen)
{
#ifdef DEBUGCHAIN
    if (!chainCleared)
    {
        DebugClear();
        chainCleared = TRUE;
    }
    DebugChar('L');
    DebugHex2(pathLen);
    DebugChar(' ');
#endif
}


//--------------------------------------------
// Trace both paths from one tile
//--------------------------------------------
int TraceSq(PlayerRec *pl, u_char *sq, int row, int col)
{
    bool    firstGood;
    int     pathLen;

#if 0
DebugClear();
DebugHex1(row);
DebugChar(':');
DebugHex1(col);
halt("-TraceSq-");
#endif
    goodPath = FALSE;       // goodpath = FALSE if already traced
    if (!(*sq & L_PATH1))   // skip if first path already traced
    {
        pathLen = TracePath(pl,sq,row,col,TRUE);
#ifdef DEBUGTRACE
DebugChar(' ');
DebugHex1(goodPath);
DebugChar('=');
DebugHex2(pathLen);
halt("-DONE1-");
#endif
        if (goodPath)       // if path is complete
        {
            AddBucket(pl,pathLen);          // record chain in a bucket
            AddScore(pl,Triangle(pathLen)); // add triangle number to score
            BoomPath(pl);                   // set boom flags if this is a good path
        }
    }
    firstGood = goodPath;

    goodPath = FALSE;       // goodpath = FALSE if already traced
    if (!(*sq & L_PATH2))   // skip if second path already traced
    {
        if (paths2[*sq & L_TILE])   // is there a second path?
        {
            pathLen = TracePath(pl,sq,row,col,FALSE);
#ifdef DEBUGTRACE
DebugChar(' ');
DebugHex1(goodPath);
DebugChar('=');
DebugHex2(pathLen);
halt("-DONE2-");
#endif
            if (goodPath)       // if path is complete
            {
                AddBucket(pl,pathLen);          // record chain in a bucket
                AddScore(pl,Triangle(pathLen)); // add triangle number to score
                BoomPath(pl);                   // set boom flags if this is a good path
            }
        }
    }

    return firstGood | goodPath;
}


//--------------------------------------------
// Trace paths from all tiles
//--------------------------------------------
int TraceAll(PlayerRec *pl)
{
    bool found;
    u_char *sq;
    int row,col;

    ClrPaths(pl);       // clear path marking flags

    found = FALSE;

    sq = pl->layout + LAYWID + 1;   // start of map data

    for (row=0; row<NROWS; row++)   // row = 0..NROWS-1
    {
        for (col=0; col<NCOLS; col++)   // col = 0..NCOLS-1
        {   // if square not empty, trace it
            if (*sq) found |= TraceSq(pl, sq, row, col);
            sq++;                   // point to next square
        }
        sq += LAYWID - NCOLS;       // go to next map line
    }

    if (found)
    {
        UpdScore(pl);
        BoomPaths(pl);  // convert marked paths to booms if any found
    }

    return found;
}


//--------------------------------------------
// Trace paths from one tile
//--------------------------------------------
int TraceOne(PlayerRec *pl, u_char *sq, int row, int col)
{
    bool found;

    ClrPaths(pl);           // clear path marking flags

    found = TraceSq(pl, sq, row, col);  // trace one square

    if (found)
    {
        UpdScore(pl);
        BoomPaths(pl);  // convert marked paths to booms if any found
    }

    return found;
}


//--------------------------------------------
// Convert surrounding squares to _BOOM1 and set next state to ST_Boom
//--------------------------------------------
void BoomBomb(PlayerRec *pl, char *sq, int row, int col)
{
    int r,c,n;
    char *p,*q;

    n = 0;
    p = sq - LAYWID - 1;
    for (r = row-1; r <= row+1; r++)
    {
        q = p;
        p += LAYWID;
        for (c = col-1; c <= col+1; c++)
        {
            if (PutBoom(pl, q, r, c)) n++;
            q++;
        }
    }

    AddScore(pl,n);
    UpdScore(pl);

    pl->state = ST_Boom;        // change next state to flash the booms
    pl->st_timr = BOOMSPEED;
    pl->st_count = NBOOMS;
}


//--------------------------------------------
// Drop the current in-air tile by one position
//--------------------------------------------
void AirDrop(PlayerRec *pl)
{
    int row,tile;
    u_char *p;

    if (pl->st_timr--) return; // delay a bit before moving
    pl->st_timr = pl->vSpeed;       // reset delay

    if (pl->joy & B_Down)           // test for down button
        pl->st_timr = 0;            // make it faster if pressing down

    row = pl->airLine - TOPLINE;    // how far from the top of the map are we?

    if (row & 0x0F)                 // are we at an even tile boundary?
    {                               // if not...
        pl->airLine += 4;           // move sprite Y down by 1/4 tile and return
        UpdSprite(pl);
        return;
    }

    row = row / 16;                 // divide by 16 to get tile row number

    // here when about to move from an even tile boundary
    // need to check for bottom or a tile below
    // AIRCOL is map column, D0 is next row

    p = pl->layout + 1 + row*LAYWID + pl->airCol; // get address of cell under sprite

    if (row < NROWS && *(p+LAYWID)==_BLANK)     // if we are not at bottom row
    {                                           // and next row's cell is empty
        pl->airLine += 4;           // move sprite Y down by 1/4 tile and return
        UpdSprite(pl);
        return;
    }

    if (row == 0)                       // are we stuck at the top of the screen?
    {
        pl->state = ST_Over;            // game over if stuck at top of screen
        return;
    }

    // here when at bottom of map or stopped by a tile

    tile = pl->airNum;                  // get new tile type
    *p = tile;                          // store tile in map
    UpdTile(pl,tile,row-1,pl->airCol,0); // update single tile on screen

    if (tile == _BOMB)                  // was the current tile a bomb?
    {                                   // if so, convert surrounding tiles to
        BoomBomb(pl,p,row-1,pl->airCol);  // bombs and set next state to ST_Boom
    }

    pl->airNum = pl->nxtNum;            // copy next tile number
    if (pl->airNum == _BOMB)            // was it a bomb?
    {
        pl->bombs--;                    // remove it from the bombs remaining
        UpdScore(pl);
    }

    pl->nxtNum  = RandTile();           // get a random tile
    pl->nxtNum2 = pl->nxtNum;           // and save it as next tile

    if (tile != _BOMB)                  // don't trace paths if it was a bomb
    {
        if (!TraceOne(pl,p,row-1,pl->airCol))
        {
//          AddScore(pl,1);             // just something to indicate the event
//          UpdScore(pl);
        }
    }

    pl->airLine = TOPLINE;              // reset to top of screen
    pl->airCol = CENTER;                // reset horiz position to center
    UpdSprite(pl);
}


#ifdef SHOWLIMIT
//--------------------------------------------
// Handle one player's time limit
// This should be called once per frame
// Returns TRUE if timer expired
//--------------------------------------------
bool DoTimeLimit(PlayerRec *pl)
{
    // check if timer already expired
    if (pl->limit == 0) return TRUE;

    // count a tick
    if (++pl->limitTick == HERTZ)
    {
        // count a second
        pl->limitTick = 0;
        pl->limit--;
        UpdScore(pl);               // update timer in score display
        return (pl->limit == 0);    // return TRUE if timer expired
    }

    // timer still running
    return FALSE;
}
#endif


//--------------------------------------------
// Handle one player's current state
//--------------------------------------------
void DoPlayer(PlayerRec *pl)
{
    RdJoy(pl);

    switch(pl->state)
    {
        case ST_InitLevel: // initializing level
            UpdSprite(pl);   // update sprites
            UpdScore(pl);    // update score/bombs
            DrawMap(pl);     // redraw entire map
            if (pl->p2Flag && !twoPlayers)
                pl->state = ST_NoPlayer;    // not playing if player 2 with only 1 player
            else
                pl->state = ST_Player;      // else player is playing
            pl->st_timr = pl->vSpeed;       // set up timer for ST_Player
            break;

        case ST_Player: // normal gameplay
#ifdef SHOWLIMIT
            if (DoTimeLimit(pl))            // check for time limit expired
            {
                pl-> state = ST_Over;       // GAME OVER if timer expired
                break;
            }
#endif
            DoPause(pl);                    // check for pause button
            DoLeftRight(pl);                // handle left/right movement
            DoButtons(pl);                  // handle up/down/C
            AirDrop(pl);                    // move the current tile down a bit
            break;

        case ST_NoPlayer: // player not playing (never started or game over)
            if (pl->p2Flag) break;          // exit if player 2
            if (P2.state == ST_NoPlayer)    // is player 2 also in this state?
            {
                pl->state   = ST_AllOver;   // move to final game over state
                pl->st_timr = GAMEOVERTIME; // final game over message delay
            }
            break;

        case ST_Over: // display a GAME OVER message and go to NoPlayer
            *VDP_CTRL_L = NAMEA_W + (pl->mapBase<<16) + NAMEPOS(9,2);
            ScrnMsg(" GAME  OVER ");
            pl->state = ST_NoPlayer;
            break;

        case ST_AllOver: // GAME OVER has been displayed
            if (pl->st_timr--) break;
            gameOver = TRUE;
            break;

        case ST_Boom:        // tiles exploding
        {
            u_char *p;
            int tile,row,col;

            if (pl->st_timr--) break;       // wait for boom timer to expire
            pl->st_timr = BOOMSPEED;        // reset boom timer
            pl->st_count--;                 // count down number of booms

            p = pl->layout + LAYWID + 1;    // start of map data
            for (row=0; row<NROWS; row++)   // row = 0 .. NROWS-1
            {
                for (col=0; col<NCOLS; col++)   // col = 0 .. NCOLS-1
                {
                    tile = *p;              // get current tile
                    if (tile == _BOOM1 || tile == _BOOM2)
                    {
                        tile ^= 1;          // swap betweeen _BOOM1 and _BOOM2
                        if (!pl->st_count)  // if this is the last boom,
                            tile = _BLANK;  // put blanks in map
                        *p = tile;          // update map;
                        UpdTile(pl,tile,row,col,0);
                    }
                    p++;            // point to next square
                }
                p += LAYWID - NCOLS;        // go to next map line
            }

        if (!pl->st_count)      // is this the last boom?
 #if 1
            pl->state   = ST_Fall;      // go back to player-is-playing
            pl->st_timr = FALLSPEED;    // set up timer for GAME_Player
 #else
            pl->state   = ST_Player;    // go back to player-is-playing
            pl->st_timr = pl->vSpeed;   // set up timer for GAME_Player
 #endif
            break;
        }

        case ST_Fall:        // tiles falling
        {
            u_char *p;
            int row,col,tile,fell;

        // FIXME: need to support half rows, maybe with Fall1/Fall2 states?
            if (pl->st_timr--) break;   // wait for timer to expire
            pl->st_timr = FALLSPEED;    // reset timer for futher falling

            p = pl->layout + LAYWID*NROWS + 1;
            fell = 0;                   // clear "something fell" flag
            for (row=NROWS-1; row; row--)       // row = NROWS-1 .. 1
            {
                for (col=0; col<NCOLS; col++)   // col = 0 .. NCOLS-1
                {   // if tile is empty, but tile above is not
                    if (*p == 0 && *(p-LAYWID) != 0)
                    {
                        tile = *(p-LAYWID);     // get falling tile
                        *p = tile;              // put falling tile in new location
                        *(p-LAYWID) = _BLANK;   // remove falling tile from old location
                        UpdTile(pl,tile  ,row  ,col,0); // draw tile in new position
                        UpdTile(pl,_BLANK,row-1,col,0); // clear old tile position
                        fell = TRUE;            // set "something fell" flag
                    }
                    p++;                // go to next map cell
                }
                p -= LAYWID+NCOLS;      // go to previous map line
            }
            if (!fell)
            {
                if (!TraceAll(pl))      // if path not found
                {
                    pl->state   = ST_Player;    // go back to player-is-playing
                    pl->st_timr = pl->vSpeed;   // set up timer for GAME_Player
                }
            }
            break;
        }

        default:
            break;
    }
}


//--------------------------------------------
// Main in-game loop, check for pause, and do each player if not paused
//--------------------------------------------
void MainLoop(void)
{
    bool    unPause;

    while (!gameOver)
    {
#if 0
*VDP_CTRL_L = NAMEA_W;
*VDP_DATA = '0' + P1.state;
*VDP_DATA = '0' + P2.state;
*VDP_DATA = frameCount & 0x7F;
//*VDP_DATA = '0' + pauseFlag;
*VDP_DATA = '0' + P1.airLine/100;
*VDP_DATA = '0' + P1.airLine/10 % 10;
*VDP_DATA = '0' + P1.airLine % 10;
#endif
#if 1
*VDP_CTRL_L = NAMEA_W;
hex2(maxStack);
#endif
        WaitVBlank();
        if (pauseFlag)
        {
            unPause = FALSE;
            RdJoy(&P1);
            if (P1.joyOn & B_Start) unPause = TRUE;
            if (twoPlayers)
            {
                RdJoy(&P2);
                if (P2.joyOn & B_Start) unPause = TRUE;
            }
            if (unPause)
            {
                pauseFlag = FALSE;
                if (P1.state != ST_NoPlayer) DrawMap(&P1);
                if (P2.state != ST_NoPlayer) DrawMap(&P2);
            }
        }
        else
        {
#ifdef DEBUGCHAIN
chainCleared = FALSE;
#endif
            DoPlayer(&P1);
            DoPlayer(&P2);
        }
    }

    if (P1.score > hiScore)
        hiScore = P1.score;

    if (twoPlayers && P2.score > hiScore)
        hiScore = P2.score;
}


//--------------------------------------------
// Display caret on proper line to show number of players
//--------------------------------------------
void TitleNPlayers(void)
{
    char ch;
    if (twoPlayers) ch = BLANK;
               else ch = CARET;
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(20,15);
    *VDP_DATA = ch;
    if (twoPlayers) ch = CARET;
               else ch = BLANK;
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(21,15);
    *VDP_DATA = ch;
}


//--------------------------------------------
// Handle player input during title screen
//--------------------------------------------
void TitlePlayer(PlayerRec *pl)
{
    RdJoy(pl);
    if (pl->joyOn & B_Start) // start button
    {
        if (twoPlayers || !pl->p2Flag)
            titleExit = TRUE;
    }
    if (pl->joyOn & B_Up)   // up
    {
        twoPlayers = FALSE; // 1 player
        TitleNPlayers();
    }
    if (pl->joyOn & B_Down) // down
    {
        twoPlayers = TRUE; // 2 players
        TitleNPlayers();
    }
}


//--------------------------------------------
// Display logo for title screen
//--------------------------------------------
#define LOGOWID     22
#define LOGOHGT     (sizeof logo/LOGOWID)

const u_char logo[][LOGOWID] =
{
  { TOPLFT_C,HORIZ_UL,HORIZ_UR,T_DN_UL,T_DN_UR,TRNDL_UL,TRNDL_UR,TRNDR_UL,TRNDR_UR,T_DN_UL,T_DN_UR,
    TRNDL_UL,TRNDL_UR,TRNDR_UL,TRNDR_UR,CAP_L_UL,CAP_L_UR,TRNDR_UL,TRNDR_UR,HORIZ_UL,HORIZ_UR,TOPRGT_C },
  { LFT_C_L,HORIZ_LL,HORIZ_LR,T_DN_LL,T_DN_LR,TRNDL_LL,TRNDL_LR,TRNDR_LL,TRNDR_LR,T_DN_LL,T_DN_LR,
    TRNDL_LL,TRNDL_LR,TRNDR_LL,TRNDR_LR,CAP_L_LL,CAP_L_LR,TRNDR_LL,TRNDR_LR,HORIZ_LL,HORIZ_LR,RGT_C_L },
  { SIDE,PLAIN_UL,PLAIN_UR,VERT_UL,VERT_UR,VERT_UL,VERT_UR,VERT_UL,VERT_UR,T_RGT_UL,T_RGT_UR,
    T_LFT_UL,T_LFT_UR,T_RGT_UL,T_RGT_UR,CAP_L_UL,CAP_L_UR,TRNUR_UL,TRNUR_UR,TRNDL_UL,TRNDL_UR,SIDE },
  { SIDE,PLAIN_LL,PLAIN_LR,VERT_LL,VERT_LR,VERT_LL,VERT_LR,VERT_LL,VERT_LR,T_RGT_LL,T_RGT_LR,
    T_LFT_LL,T_LFT_LR,T_RGT_LL,T_RGT_LR,CAP_L_LL,CAP_L_LR,TRNUR_LL,TRNUR_LR,TRNDL_LL,TRNDL_LR,SIDE },
  { SIDE,BOMB_UL,BOMB_UR,CAP_U_UL,CAP_U_UR,TRNUR_UL,TRNUR_UR,UL_DR_UL,UL_DR_UR,T_UP_UL,T_UP_UR,
    TRNUL_UL,TRNUL_UR,TRNUR_UL,TRNUR_UR,HORIZ_UL,HORIZ_UR,HORIZ_UL,HORIZ_UR,TRNUL_UL,TRNUL_UR,SIDE },
  { SIDE,BOMB_LL,BOMB_LR,CAP_U_LL,CAP_U_LR,TRNUR_LL,TRNUR_LR,UL_DR_LL,UL_DR_LR,T_UP_LL,T_UP_LR,
    TRNUL_LL,TRNUL_LR,TRNUR_LL,TRNUR_LR,HORIZ_LL,HORIZ_LR,HORIZ_LL,HORIZ_LR,TRNUL_LL,TRNUL_LR,SIDE },
  { BOTLFT,BOT,BOT,BOT,BOT,BOT,BOT,BOT_C_L,BOT_C_R,BOT,BOT,
    BOT,BOT,BOT,BOT,BOT,BOT,BOT,BOT,BOT,BOT,BOTRGT },
};

void ShowLogo(void)
{
    int i,j;

    for (i=0; i<LOGOHGT; i++)
    {
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(i+6,9);
        for (j=0; j<LOGOWID; j++)
            *VDP_DATA = logo[i][j];
    }
}


//--------------------------------------------
// Title screen
//--------------------------------------------
void Title(void)
{
    int i;
    extern char _VersionID[];

    ClearSprites();
    ClearScreen();
    ShowLogo();
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(17,(40-20)/2);
    ScrnMsg("Press START to begin");
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(24,(40-30)/2);
    ScrnMsg("Copyright 2009 by Bruce Tomlin");
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(20,17);
    ScrnMsg("1 PLAYER");
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(21,17);
    ScrnMsg("2 PLAYERS");
    *VDP_CTRL_L = NAMEA_W + NAMEPOS(27,11);
    for (i=0; i<18; i++)
        *VDP_DATA = _VersionID[i];
    TitleNPlayers();

    if (1||hiScore)
    {
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(2,20-8/2);
        ScrnMsg("HI-SCORE");

        sprintf(buffer,"%06lu",hiScore);
        *VDP_CTRL_L = NAMEA_W + NAMEPOS(3,20-6/2);
        ScrnMsg(buffer);
    }

    titleExit = FALSE;
    while (!titleExit)
    {
        WaitVBlank();
        TitlePlayer(&P1);
        TitlePlayer(&P2);
    }
}


//--------------------------------------------
// Outer game loop, do a title scren then play a game
//--------------------------------------------
void GameLoop(void)
{
    while(1)
    {
        Title();
        RandTile();
        ClearScreen();
#if 1
        Randomize();
#endif
        InitPlayers();
        InitSprites();
#if 0
P1.layout[0x49] = _CURVE+2; P1.layout[0x4A] = _CROSS;     P1.layout[0x4B] = _CURVE+3;
P1.layout[0x52] = _CURVE+1; P1.layout[0x53] = _CROSSOVER; P1.layout[0x54] = _CURVE;
P1.layout[0x5B] = _CAP+3;   P1.layout[0x5C] = _TJOINT+2;  P1.layout[0x5D] = _CAP+1;
#endif
#if 0
int row,col;
for (row=1; row<=NROWS; row++)
    for (col=0; col<=NCOLS; col++)
        P1.layout[row*LAYWID+LAYWID+col+1] = _CROSS;
#endif
        gameOver = FALSE;
        MainLoop();
    }
}


//--------------------------------------------
// Initialize system and start outer game loop
//--------------------------------------------
int main(void)
{
    InitCRAM();
    init_font();
    init_tiles();
    InitJoy();
    ClearScreen();

//    *VDP_CTRL = 0x8144; // C00004 reg 1 = 0x44 unblank display
//    asm volatile("move.w #0x2700,%sr");     // disenable vsync interrupt

//    *VDP_CTRL = 0x8164; // C00004 reg 1 = 0x64 unblank display, enable vsync int
//    asm volatile("move.w #0x2500,%sr");     // enable vsync interrupt

    asm volatile("move.w #0x2500,%sr");     // enable vsync interrupt
    *VDP_CTRL = 0x8134; // C00004 reg 1 = 0x34 enable V-sync interrupt and DMA
*VDP_CTRL = 0x8174; // C00004 reg 1 = 0x74 unblank display, enable V-sync interrupt and DMA

    hiScore = 0;
    Randomize();
    InitPlayers();
    InitSprites();
//  twoPlayers = TRUE;     // default to 2 players
    twoPlayers = FALSE;    // default to 1 player
    GameLoop();

    return 0;
}
